import gymnasium as gym
from gymnasium import Env
from gymnasium.spaces import Box
import numpy as np
import importlib
import random
from matplotlib import pyplot as plt
from typing import Any, Tuple, Optional, Dict

from rlnav.point_maze.utils.indexes import TileType, Colors, PointMazeMapsIndex


class PointMazeV0(Env):
    """
    2D Point Maze Navigation Environment.
    In the code below, x/y refers to the agent's position in the observation space, i/j refers to the agent's position
    in the maze grid (i refers to the row id, j to the column id).
    """

    name = "Point-Maze"
    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 10}

    def __init__(self, **params):
        """
        Initialize the Point Maze environment.

        Args:
            map_name: Name of the map module to load (default: "EMPTY")
            action_noise: Standard deviation of noise added to actions (default: 1.0)
            reset_anywhere: Whether to reset agent at any valid position (default: True)
            goal_conditioned: Whether to use goal-conditioned RL (default: False)
            render_mode: Rendering mode (default: "rgb_array")
        """
        self.map_name: str = params.get("map_name", PointMazeMapsIndex.EMPTY.value)
        self.action_noise = params.get("action_noise", 1.0)
        self.reset_anywhere = params.get("reset_anywhere", True)
        self.goal_conditioned = params.get("goal_conditioned", False)
        self.render_mode = params.get("render_mode", "rgb_array")
        self.render_resolution = params.get("render_resolution", 10)  # Pixels per grid cell
        if isinstance(self.action_noise, int):
            self.action_noise = float(self.action_noise)
        assert isinstance(self.action_noise, float) and self.action_noise >= 0, "Invalid action_noise value."
        assert isinstance(self.reset_anywhere, bool), "reset_anywhere must be a boolean."
        assert isinstance(self.goal_conditioned, bool), "goal_conditioned must be a boolean."
        assert self.render_mode in self.metadata["render_modes"], f"Invalid render_mode: {self.render_mode}"

        # Load the maze map from the given map name
        module_path = f"rlnav.point_maze.maps.{self.map_name}"
        try:
            self.maze_array = np.array(importlib.import_module(module_path).maze_array, dtype=np.float16)
        except (ImportError, AttributeError) as e:
            raise ValueError(f"Failed to load maze map '{self.map_name}': {e}")

        self.height, self.width = self.maze_array.shape

        # Define the observation and action spaces
        self.observation_space: Box = Box(
            low=np.array([-self.width / 2, -self.height / 2], dtype=np.float32),
            high=np.array([self.width / 2, self.height / 2], dtype=np.float32)
        )
        self.action_space: Box = Box(
            low=np.array([-1.0, -1.0], dtype=np.float32),
            high=np.array([1.0, 1.0], dtype=np.float32)
        )

        # Initialize state variables
        self.agent_observation = None
        self.goal = None

        # For rendering
        self.window = None
        self.clock = None

        # Reset to initialize agent position and goal
        self.reset()

    def _sample_reachable_position(self) -> np.ndarray:
        """Sample a random reachable position in the maze."""
        empty_tiles = np.argwhere(
            np.logical_or(self.maze_array == TileType.EMPTY.value,
                          self.maze_array == TileType.REWARD.value))
        if len(empty_tiles) == 0:
            raise ValueError("Looking for a reachable position but none was found.")

        # Sample a reachable tile
        reachable_tile = random.choice(empty_tiles)
        # Reachable position at the center of the tile
        position = self.get_observation(*reachable_tile)
        # Sample a point in the selected tile, avoid np.random to keep the reset seed for random package
        noise = np.array([random.random(), random.random()]) - 0.5
        return position + noise

    def get_observation(self, i: int, j: int) -> np.ndarray:
        """Converts grid coordinates to an observation that belongs to the center of the tile."""
        return np.array([j + 0.5 - self.width / 2, -(i + 0.5 - self.height / 2)])

    def get_coordinates(self, observation: np.ndarray) -> Tuple[int, int]:
        """Return the tile that belongs to the given observation."""
        i = int(- observation[1] + self.height / 2)
        j = int(observation[0] + self.width / 2)

        # Ensure coordinates are within bounds
        i = max(0, min(i, self.height - 1))
        j = max(0, min(j, self.width - 1))

        return i, j

    def reset(self, *,
              seed: Optional[int] = None,
              options: Optional[Dict[str, Any]] = None) -> Tuple[np.ndarray, Dict]:
        """
        Reset the environment to an initial state.

        Args:
            seed: Random seed for reproducibility
            options: Additional options for reset customization

        Returns:
            Initial observation and info dict
        """
        # Set seeds if provided
        if seed is not None:
            random.seed(seed)
            np.random.seed(seed)
            self.np_random, seed = gym.utils.seeding.np_random(seed)

        # Reset the agent's position
        if self.reset_anywhere:
            self.agent_observation = self._sample_reachable_position()
        else:
            valid_tiles = np.argwhere(self.maze_array == TileType.START.value)
            if len(valid_tiles) == 0:
                raise ValueError("Cannot reset with reset_anywhere=False and no available start tiles.")
            self.agent_observation = self.get_observation(*random.choice(valid_tiles))

        assert self.observation_space.contains(self.agent_observation.astype(self.observation_space.dtype)), \
            f"Invalid observation after reset: {self.agent_observation}"

        # Sample a goal for goal-conditioned RL
        info = {}
        if self.goal_conditioned:
            self.goal = self._sample_reachable_position()
            info["goal"] = self.goal.copy()

        return self.agent_observation.copy(), info

    def is_available(self, i: int, j: int) -> bool:
        """
        Check if a position (i, j) in the maze array is available (not a wall or out of bounds).

        Args:
            i: Row index
            j: Column index

        Returns:
            True if the position is available, False otherwise
        """
        return (0 <= j < self.width and
                0 <= i < self.height and
                self.maze_array[i, j] != TileType.WALL.value)

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, float, bool, bool, Dict]:
        """
        Move the agent according to the action and return the new state.

        Args:
            action: Action to take

        Returns:
            Tuple of (observation, reward, terminated, truncated, info)
        """
        assert self.action_space.contains(action.astype(self.action_space.dtype)), \
            f"Invalid action: {action}"

        # Add action noise
        action = np.clip(action + np.random.normal(0, self.action_noise, size=action.shape),
                         self.action_space.low, self.action_space.high)

        # Implement sub-stepping for smoother movement
        for _ in range(10):  # Sub-steps
            new_observation = self.agent_observation + action / 10
            if not self.is_available(*self.get_coordinates(new_observation)):
                break
            self.agent_observation = new_observation

        # Get current tile type
        i, j = self.get_coordinates(self.agent_observation)
        tile_type = self.maze_array[i, j]

        # Calculate reward and check termination conditions
        reward = 10.0 if tile_type == TileType.REWARD.value else -0.1

        # Check if agent reached trap
        terminated = tile_type == TileType.TRAP.value
        truncated = False

        # Additional info for debugging
        info = {
            "position": self.agent_observation.copy(),
            "coordinates": (i, j),
            "tile_type": int(tile_type)
        }

        # If goal-conditioned, add distance to goal in info
        if self.goal_conditioned:
            distance_to_goal = np.linalg.norm(self.agent_observation - self.goal)
            info["distance_to_goal"] = distance_to_goal
            info["goal"] = self.goal.copy()

            # Add goal-based reward component
            if distance_to_goal < 0.5:  # Within half a tile of goal
                reward += 5.0
                terminated = True
        return self.agent_observation.copy(), reward, terminated, truncated, info

    def render(self, show_agent=True, show_rewards=True, show_grid=False):
        """
        Render the maze environment with continuous agent position.

        Returns:
            RGB array representation of the environment
        """
        # Create a blank canvas
        img = np.zeros((self.height * self.render_resolution, 
                        self.width * self.render_resolution, 3), dtype=np.uint8)

        # Draw the maze grid
        for i in range(self.height):
            for j in range(self.width):
                tile_type = int(self.maze_array[i, j])
                y_start = i * self.render_resolution
                y_end = (i + 1) * self.render_resolution
                x_start = j * self.render_resolution
                x_end = (j + 1) * self.render_resolution

                # Set color based on tile type
                if tile_type == TileType.EMPTY.value:
                    color = Colors.EMPTY.value
                elif tile_type == TileType.WALL.value:
                    color = Colors.WALL.value
                elif tile_type == TileType.START.value:
                    color = Colors.START.value
                elif tile_type == TileType.REWARD.value:
                    color = Colors.REWARD.value
                elif tile_type == TileType.TRAP.value:
                    color = Colors.TRAP.value
                else:
                    color = Colors.EMPTY.value

                img[y_start:y_end, x_start:x_end] = color

                # Add tile border
                if show_grid and tile_type != TileType.WALL.value:
                    img[y_start:y_start + 1, x_start:x_end] = Colors.TILE_BORDER.value  # Top
                    img[y_end - 1:y_end, x_start:x_end] = Colors.TILE_BORDER.value  # Bottom
                    img[y_start:y_end, x_start:x_start + 1] = Colors.TILE_BORDER.value  # Left
                    img[y_start:y_end, x_end - 1:x_end] = Colors.TILE_BORDER.value  # Right

        # Draw the agent using continuous coordinates
        # Assuming self.agent_x and self.agent_y are the continuous coordinates (bounded)
        # and self.x_bounds = (x_min, x_max), self.y_bounds = (y_min, y_max) define the bounds

        if show_agent:
            img = self.place_point(img, Colors.AGENT.value, *self.agent_observation)

        # Draw the goal if goal-conditioned and with continuous coordinates
        if show_rewards and self.goal_conditioned and hasattr(self, 'goal_x') and hasattr(self, 'goal_y'):
            img = self.place_point(img, Colors.GOAL.value, self.goal_x, self.goal_y)

        if self.render_mode == "human":
            # Display the image using matplotlib
            plt.figure(figsize=(8, 8))
            plt.imshow(img)
            plt.axis('off')
            plt.title(f"{self.name} Environment")
            plt.show()

        return img

    def place_point(self, image: np.ndarray, color, position_x, position_y, radius: Optional[float] = None) -> np.ndarray:
        x_min, y_min = self.observation_space.low
        x_max, y_max = self.observation_space.high
        img_height, img_width = image.shape[:2]

        # Normalize the agent's position to the image dimensions
        x_norm = (position_x - x_min) / (x_max - x_min) * img_width
        y_norm = (position_y - y_min) / (y_max - y_min) * img_height

        x_center = int(x_norm)
        y_center = int(y_norm)
        if not radius:
            radius = self.render_resolution // 3

        # Draw a circle for the agent
        y, x = np.ogrid[:img_height, :img_width]
        dist = np.sqrt((y - y_center) ** 2 + (x - x_center) ** 2)
        mask = dist <= radius
        image[mask] = color
        return image

    def close(self):
        """Clean up resources."""
        if self.window is not None:
            plt.close('all')
            self.window = None

    def copy(self):
        """Create a copy of this environment."""
        return PointMaze(
            map_name=self.map_name,
            action_noise=self.action_noise,
            reset_anywhere=self.reset_anywhere,
            goal_conditioned=self.goal_conditioned,
            render_mode=self.render_mode,
            render_resolution=self.render_resolution
        )


from maps_index import MapsIndex
from tile_type import TileType
import importlib


def build_map(map_name, corridors_width=3):
    if isinstance(map_name, MapsIndex):
        map_name = map_name.value
    maze_map = importlib.import_module("rlnav.maps." + map_name).maze_array

    # Find rows that can be duplicated
    rows_to_duplicate = []
    for index, row in enumerate(maze_map):
        horizontal_wall_length = 0
        for cell in row:
            if cell == TileType.WALL.value:
                horizontal_wall_length += 1
            if horizontal_wall_length > 1:
                # This row have a horizontal wall. Skip it
                break
        else:
            # This row have no horizontal wall. Save it.
            rows_to_duplicate.append(index)
    del horizontal_wall_length

    # Find columns that can be duplicated
    columns_to_duplicate = []
    for column_index in range(len(maze_map[0])):
        vertical_wall_length = 0
        for cell in row:
            if cell == TileType.WALL.value:
                vertical_wall_length += 1
            else:
                vertical_wall_length = 0
            if vertical_wall_length > 1:
                # This row have a horizontal wall. Skip it
                print("skipped row ", column_index, " because of a too long wall at position ", column_index, sep="")
                vertical_wall_length = 0
                break
        else:
            # This row have no horizontal wall. Save it.
            print("Added row ", index, sep="")
            columns_to_duplicate.append(column_index)

    # Duplicate the rows and columns
    new_maze_array = []
    for index, row in enumerate(maze_map):
        new_maze_array += [row] * (corridors_width if index in rows_to_duplicate else 1)
        # if index in columns_to_duplicate:
        #     new_maze_array += [[row]]
        # else:
        #     new_maze_array += [[row]] * corridors_width

    for column_index in range(len(maze_map[0])):
        if column_index in columns_to_duplicate:
            for row_index in range(len(maze_array[0])):
                element_to_add = maze_array[row_index][column_index]

if __name__ == "__main__":
    build_map(MapsIndex.FOUR_ROOMS, corridors_width=4)



#!/usr/bin/env python
from __future__ import print_function
from ...utils import generate_video
import argparse
from time import sleep
from vizdoom import *
import os
import numpy as np

parser = argparse.ArgumentParser()
parser.add_argument('map_name', default="medium")
parser.add_argument('-fs', '--fixed_start', default=True)
flags = parser.parse_args()

if __name__ == '__main__':
    current_directory = os.path.dirname(os.path.abspath(__file__))
    wads_directory = current_directory + "/wads/" + ("fixed_start" if flags.fixed_start else "reset_anywhere")
    wad_directory_path = wads_directory + "/" + flags.map_name + ".wad"

    game = DoomGame()
    game.load_config("./configurations/default.cfg")
    game.set_doom_scenario_path(wad_directory_path)
    game.add_game_args("+freelook 1")
    game.set_screen_resolution(ScreenResolution.RES_640X480)
    game.set_window_visible(True)
    game.set_doom_map("MAP00")
    game.set_mode(Mode.SPECTATOR)
    game.init()
    episodes = 10

    for i in range(episodes):
        print("Episode #" + str(i + 1))
        images = []

        game.new_episode()
        while not game.is_episode_finished():
            state: GameState = game.get_state()
            images.append(np.moveaxis(state.screen_buffer.copy(), 0, -1))
            game.advance_action()
            last_action = game.get_last_action()
            reward = game.get_last_reward()
        generate_video(images, current_directory, "episode_" + str(i))

        print("Episode finished!")
        print("Total reward:", game.get_total_reward())
        print("************************")
        sleep(2.0)

    game.close()



import importlib
import math
import os
import random
from enum import Enum
from tempfile import TemporaryDirectory
from typing import Union, Tuple
from skimage.draw import line_aa
from scipy.spatial import distance
import numpy as np
from omg import WAD, MapEditor
from vizdoom import DoomGame, ScreenResolution, Mode, Gameobservation
from gymnasium.spaces import Box, Discrete
from .maps.maps_index import MapsIndex
from .maps.tile_type import TileType

from ...utils import generate_video, save_image
from .wads_builder import build_wad


class Colors(Enum):
    EMPTY = [250, 250, 250]
    WALL = [50, 54, 51]
    START = [213, 219, 214]
    TERMINAL = [73, 179, 101]
    TILE_BORDER = [50, 54, 51]
    AGENT = [0, 0, 255]
    GOAL = [255, 0, 0]


class DoomMaze(Env):

    name = "VizDoom-Maze"

    def __init__(self,
                 map_tag: MapsIndex = MapsIndex.MEDIUM,
                 reset_anywhere: bool = False,
                 random_walls_textures: bool = True,
                 sparse_reward: bool = True,
                 discrete_goals: bool = False,
                 goal_range: Union[Tuple[int, int], None] = None,
                 behaviour_file: str = "static_goal.acs",
                 configuration_file: str = "default.cfg",
                 block_size: int = 96,
                 attainability_threshold: int = 64,
                 pixels_per_tile: int = 20,
                 pixels_wall_width: int = 3,
                 action_frame_repeat: int = 4,
                 observation_resolution: ScreenResolution = ScreenResolution.RES_160X120):
        """
        Instantiate the class and build the .wad file.

        CONVENTION: There is two type of coordinates used in this file.
            - The coordinates of a tile inside the map_array (the array that describe the maze shape, a 2D list of
            TileType values (integers)). These coordinates are elements of a discrete space and are identified with
            h (for height) and w (width). self.map_array[h, w] return an int that represent the type of the tile at
            coordinates (h, w).
            The keyword "tile" in variable names refers to a tile in self.map_array so "tile_coordinates" refers to a
            coordinate (h, w) in the map_array.
            - The coordinates of a position inside the VizDoom maze, identified using x and y.
            These values are continuous.
             -> Functions self.get_tile_from_position and self.get_position_from_tile help doing the bridge from one to
             another.

        :param map_tag: element of DoomMapIndex enum that indicate from which map the .wad file should be built.
        :param reset_anywhere: If true, the agent position is sampled uniformly in the maze at each episode start.
            Otherwise, it is always set to one of the tiles equal to environments.maps.tile_typeTileType.START.value.
        :param random_walls_textures: Boolean, Indicate whether walls textures should be sampled randomly or not.
        :param sparse_reward: If true, the agent receive a reward of 0 when it reached the goal, -1 otherwise.
            If false, it will receive a reward equal to minus the Euclidean distance between its position and the goal
            one.
        :param discrete_goals: Boolean, Indicate whether goals should be sampled from a pre-defined set of fixed goals
            or uniformly in the entire maze
        :param goal_range: Union[Tuple[int, int], None]: Maximum distance allowed between the initial observation
            tile coordinates and the sampled goal position.
        :param behaviour_file: str: Path to the behaviour file. Should be a .cfg file. A default file is available for
            more information.
        :param configuration_file: str: Path to the configuration file. Should be a .acs file. A default file is
            available for more information.
        :param block_size: Size of a self.maze_map tile in the vizdoom maze (in pre-set maps, this is also the corridor
            width).
        :param attainability_threshold: We will consider a goal reached if the distance between the agent and the goal
            is bellow this threshold.
        :param pixels_per_tile: Only used for the self.render(mode="top_view"). Indicate the number of pixels for each
            self.maze_map tiles.
        :param pixels_wall_width: Only used for the self.render(mode="top_view"). Indicate the walls width in pixels
            when a wall is drawn.
        :param action_frame_repeat: How many times an action selected by the agent will be performed in the environment
            before we return the next observation, the reward, and the 'done' boolean.
        :param observation_resolution: Resolution (in pixels) of both the observation and the goal image given to the
            agent. Other values available: ScreenResolution.RES_320X240, ScreenResolution.RES_640X480.
        """
        assert isinstance(map_tag, MapsIndex), "Map tag should be an instance of .maps.maps_index.MapsIndex"
        assert isinstance(reset_anywhere, bool)
        assert isinstance(random_walls_textures, bool)
        assert isinstance(sparse_reward, bool)
        assert isinstance(discrete_goals, bool)
        assert not discrete_goals or goal_range is None, \
            "Impossible to ensure goals are close to the agent if goals are fixed."
        # '-> Moving the agent initial position closed to a fixed goal is not implemented.
        assert goal_range is None or 0 <= goal_range[0] <= goal_range[1]
        if "." not in behaviour_file:
            behaviour_file += ".acs"
        if "." not in configuration_file:
            configuration_file += ".cfg"
        assert behaviour_file.endswith(".acs"), "Wrong behaviour file extension, expected '.acs'"
        assert configuration_file.endswith(".cfg"), "Wrong behaviour file extension, expected '.cfg'"
        config_files_directory_path = os.path.dirname(os.path.abspath(__file__)) + "/configurations/"
        assert os.path.exists(config_files_directory_path + behaviour_file), \
            "Behaviour file not found. Verify that this file exists in " + config_files_directory_path + "."
        assert os.path.exists(config_files_directory_path + configuration_file), \
            "Behaviour file not found. Verify that this file exists in " + config_files_directory_path + "."
        assert isinstance(block_size, int) and block_size > 0
        assert isinstance(pixels_per_tile, int) and pixels_per_tile > 0
        assert isinstance(pixels_wall_width, int) and pixels_per_tile >= pixels_wall_width > 0
        assert isinstance(action_frame_repeat, int) and 0 < action_frame_repeat

        self.map_tag = map_tag
        self.map_name = map_tag.name.lower()
        self.reset_anywhere = reset_anywhere
        self.random_walls_textures = random_walls_textures
        self.sparse_reward = sparse_reward
        self.behaviour_file = behaviour_file
        self.configuration_file = configuration_file
        self.block_size = block_size
        self.attainability_threshold = attainability_threshold
        self.discrete_goals = discrete_goals
        self.goal_range = goal_range
        self.pixels_per_tile = pixels_per_tile
        self.pixels_wall_width = pixels_wall_width
        self.action_frame_repeat = action_frame_repeat
        self.observation_resolution = observation_resolution

        # Generate wad file and save it in a temp directory (prevent concurrent access if two
        # environment are created at the same time on the same wad file)
        self.wad_temporary_directory = TemporaryDirectory()
        self.map_array = np.array(
            importlib.import_module("environments.doom_visual_navigation.maps." + self.map_name).maze_array)
        print("Building wad file ... ", end="\r")
        build_wad(self.wad_temporary_directory.name, self.map_array,
                  random_walls_textures=self.random_walls_textures, behaviour_file=self.behaviour_file,
                  block_size=self.block_size, discrete_goals_set=self.discrete_goals)
        print("Building wad file ... DONE")

        # Build DOOM GAME !
        self.reset_game(with_wad=False)

        # Environment variables
        self.goal_tile_coordinates = None

        self.position_space = Box(low=np.zeros(2),
                                  high=np.array([self.map_array.shape[1] * self.block_size,
                                                 self.map_array.shape[0] * self.block_size]))
        available_buttons = self._game.get_available_buttons()
        self.action_space = Discrete(len(available_buttons))
        observation_shape = self.observation_resolution.name.split("_")[-1].split("X")
        observation_shape = tuple([3] + [int(elt) for elt in observation_shape])
        self.observation_space = Box(low=np.zeros(observation_shape), high=np.full(observation_shape, 255))

        # Goals
        self.goal_image = None
        if self.discrete_goals:
            self.goals_positions = list(zip(*np.where(self.map_array == TileType.TERMINAL.value)))

    @property
    def goal_env_position(self):
        return self.get_position_from_tile(*self.goal_tile_coordinates)

    def get_tile_from_position(self, x, y):
        """
        Compute the coordinates in self.map_array from the position in the viz doom maze.
        :param x: x coordinate in the viz doom maze.
        :param y: y coordinate in the viz doom maze.
        :return: tuple: Coordinates in the map_array.
        """
        h = (y - int(self.block_size / 2)) / self.block_size
        w = (x - int(self.block_size / 2)) / self.block_size
        h = len(self.map_array) - 1 - h
        return h, w

    def get_position_from_tile(self, h, w) -> (float, float):
        """
        Compute the position in the viz doom maze from the coordinates in self.map_array.
        :param w: x coordinate in the map_array.
        :param h: y coordinate in the map_array.
        :return: tuple: Position in the viz doom maze.
        """
        h = len(self.map_array) - 1 - h
        x, y = w * self.block_size, h * self.block_size
        x += int(self.block_size / 2)
        y += int(self.block_size / 2)
        return x, y

    def get_agent_position(self):
        """
        Return the agent position in the vizdoom maze.
        :return: tuple
        """
        observation = self._game.get_observation()
        agent_coordinates = np.array([observation.game_variables[0], observation.game_variables[1]])
        return agent_coordinates

    def distance_to_goal(self) -> float:
        """
        Compute the distance between the agent and the goal.
        :return: float
        """
        goal_coordinates = self.goal_env_position
        agent_coordinates = self.get_agent_position()
        return np.linalg.norm((goal_coordinates - agent_coordinates))

    def reset_game(self, with_wad=True):

        self._game = DoomGame()
        current_dir = os.path.dirname(os.path.abspath(__file__))
        self._game.set_screen_resolution(self.observation_resolution)
        self._game.load_config(current_dir + "/configurations/" + self.configuration_file)

        if with_wad:
            # Reload the map
            self._game.set_doom_scenario_path(self.wad_temporary_directory.name + "/map.wad")
            self._game.set_doom_map("MAP00")
        self._game.init()

    def get_tiles_under_max_distance(self, h, w):
        """
        Find every tile in self.map_array that are at a distance of n from the agent initial tile.
        We will use a Dijkstra-like algorithm to find the distance between every tiles and the chosen start tile.

        :param h: Tile coordinate over the map width.
        :param w: Tile coordinate over the map height.
        :return: Tiles that are close enough from the given tile coordinates.
        """

        result = []
        seen = np.full_like(self.map_array, False)
        height, width = self.map_array.shape

        def __get_neighbors(h, w):
            neighbors = []
            potential_neighbors = [(h - 1, w), (h + 1, w), (h, w - 1), (h, w + 1)]
            for h_, w_ in potential_neighbors:
                if 0 <= h_ < height and 0 <= w_ < width and self.get_tile_type(h_, w_) != TileType.WALL:
                    neighbors.append((h_, w_))
            return neighbors

        def __explore(tile_h, tile_w, distance_so_far=0):
            """
            Recursive function that, according to a given tile, and a distance from a source tile, verify if this tile
            is not too far, and if so call itself recursively on the given tile neighbors.
            Because of the usage of a shared memory list "seen", a tile cannot be added twice in the result list.

            :param tile_h: Given tile height coordinate, aka. self.map_array[tile_h, tile_w] give the right tile.
            :param tile_w: Given tile width coordinate, aka. self.map_array[tile_h, tile_w] give the right tile.
            :param distance_so_far: Distance from the original call tile and the current one.
            :return: If called with distance_so_far=0, return a list of tiles coordinates that are at a distance
            between bounds defined by self.goal_range.
            """
            assert self.goal_range is not None and 0 <= self.goal_range[0] <= self.goal_range[1]
            if seen[tile_h, tile_w] or distance_so_far > self.goal_range[1]:
                return
            if distance_so_far >= self.goal_range[0]:
                result.append((tile_h, tile_w))
            seen[tile_h, tile_w] = True

            if distance_so_far + 1 > self.goal_range[1]:
                return  # No need to go further

            for neighbor in __get_neighbors(tile_h, tile_w):
                __explore(*neighbor, distance_so_far=distance_so_far + 1)

        __explore(h, w)
        return result

    def reset(self):

        # Load wad file
        wad = WAD(from_file=os.path.join(self.wad_temporary_directory.name, "map.wad"))
        map = wad.maps["MAP00"]
        map_editor = MapEditor(map)

        # Choose agent start tile.
        if self.reset_anywhere:
            agent_start_tile_candidates = np.where(self.map_array != 1)
        else:
            agent_start_tile_candidates = np.where(self.map_array == 2)
        agent_start_tile_candidates = list(zip(*agent_start_tile_candidates))
        agent_start_tile = random.choice(agent_start_tile_candidates)

        # Choose goal tile
        if self.discrete_goals:
            self.goal_tile_coordinates = random.choice(self.goals_positions)
            goal_x, goal_y = self.get_position_from_tile(*self.goal_tile_coordinates)
        else:
            if self.goal_range is not None:
                candidates = self.get_tiles_under_max_distance(*agent_start_tile)
            else:
                candidates = list(zip(*np.where(self.map_array != 1)))
            h, w = random.choice(candidates)
            assert self.get_tile_type(h, w) != TileType.WALL
            self.goal_tile_coordinates = (h, w)
            goal_x, goal_y = self.get_position_from_tile(h, w)

            # Find the things that belong to the goal
            for thing_id, thing in enumerate(map_editor.things):
                if thing.type == 32:
                    thing.x = goal_x
                    thing.y = goal_y
                    break

        # Get goal image
        # - Place the agent next to the goal facing it
        angle = random.random() * 2 * math.pi

        agent_x = goal_x + math.cos(angle) * self.attainability_threshold
        agent_y = goal_y + math.sin(angle) * self.attainability_threshold

        # Place agent
        for thing_id, thing in enumerate(map_editor.things):
            if thing.type == 1:
                thing.x = int(agent_x)
                thing.y = int(agent_y)
                thing.angle = int(math.degrees(angle + math.pi))
                break

        # Get agent pov
        self.reset_game()  # Reload the modified wad, so we can take a picture of the goal before to replace the agent
        self.goal_image = self.observation.copy()

        # - Get agent observation and store it as a goal to reach.

        # Place agent
        x, y = self.get_position_from_tile(*agent_start_tile)
        for thing_id, thing in enumerate(map_editor.things):
            if thing.type == 1:  # A thing with a type of 1 is the agent
                thing.x = x
                thing.y = y
                thing.angle = 0  # Arbitrary looking to the right, can be changed with a value between 0 and 360

        # Save the map,
        wad.maps['MAP00'] = map_editor.to_lumps()

        # Start the episode,
        # os.system('nautilus ' + str(self.wad_temporary_directory.name))
        self._game.new_episode()
        self.reset_game()  # Reload the modified wad, so we can take a picture of the goal before to replace the agent

        return self.observation, self.goal_image

    def step(self, action=None):

        if action is None:
            self._game.advance_action()
        else:
            actions = [0 for _ in range(self.action_space.n)]
            actions[action] = 1
            self._game.make_action(actions, self.action_frame_repeat)
        distance_to_goal = self.distance_to_goal()
        reached = distance_to_goal <= self.attainability_threshold

        if self.sparse_reward:
            reward = -1 if reached else 0
        else:
            reward = - distance_to_goal

        return self.observation, reward, reached, {"reached": reached}

    @property
    def observation(self):
        """
        Return the agent observation.
        """
        observation: Gameobservation = self._game.get_observation()
        observation = observation.screen_buffer.copy()
        return observation

    def get_tile_type(self, h, w):
        return TileType(self.map_array[h, w].item())

    def get_color(self, h, w, ignore_agent=False, ignore_terminals=False):
        agent_h, agent_w = self.get_tile_from_position(*self.get_agent_position())
        if (agent_h, agent_w) == (h, w) and not ignore_agent:
            return Colors.AGENT.value
        else:
            tile_type = self.get_tile_type(h, w)
            if tile_type == TileType.START:
                return Colors.START.value
            elif tile_type == TileType.WALL:
                return Colors.WALL.value
            elif tile_type == TileType.EMPTY:
                return Colors.EMPTY.value
            elif tile_type == TileType.TERMINAL:
                return Colors.EMPTY.value if ignore_terminals else Colors.TERMINAL.value
            else:
                raise AttributeError("Unknown tile type")

    def reached(self):
        raise NotImplementedError()

    """
    Every functions bellow this are used for rendering.
    """

    def render(self, mode="top_view", ignore_goal=False, ignore_agent=False):
        assert mode in ["top_view", "observation", "observation_goal"]
        ignore_terminals = ignore_goal or not self.discrete_goals
        if mode == "top_view":
            height, width = self.map_array.shape

            # Compute the total grid size
            width_px = height * self.pixels_per_tile
            height_px = width * self.pixels_per_tile

            # Build an empty image
            img = np.full(fill_value=255, shape=(height_px, width_px, 3), dtype=np.uint8)

            # Render the grid
            for h in range(height):
                for w in range(width):
                    tile_type = self.get_tile_type(h, w)
                    position_1 = self.get_position_from_tile(h, w)

                    if tile_type == TileType.WALL:
                        neighbors = [(h + 1, w), (h, w + 1)]  # TODO supprimer les -1 et tester
                        for h_, w_ in neighbors:
                            if 0 <= h_ < height and 0 <= w_ < width and self.get_tile_type(h_, w_) == TileType.WALL:
                                position_2 = self.get_position_from_tile(h_, w_)
                                self.place_edge(img, *position_1, *position_2, Colors.WALL.value)

                        # If there is no walls next to this one, we still want to place a point (like a pillar)
                        # on this one.
                        self.place_point(img, *position_1, Colors.WALL.value, self.pixels_wall_width)
                    elif not ignore_terminals and tile_type == TileType.WALL:
                        img = self.place_point(img, *position_1, Colors.GOAL.value)
                    elif not ignore_goal and not self.discrete_goals and (h, w) == self.goal_tile_coordinates:
                        img = self.place_point(img, *position_1, Colors.GOAL.value)
            if not ignore_agent:
                img = self.place_point(img, *self.get_agent_position(), Colors.AGENT.value)
            return img

        elif mode == "observation":
            return np.moveaxis(self._game.get_observation().screen_buffer.copy(), 0, -1)
        elif mode == "observation_goal":
            observation = np.moveaxis(self._game.get_observation().screen_buffer.copy(), 0, -1)
            goal = np.moveaxis(self.goal_image.copy(), 0, -1)
            return np.concatenate((observation, goal), 1)

    def place_point(self, image: np.ndarray, position_x, position_y, color: Union[np.ndarray, list], width=5):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the observation space of the point to place.
        param y: y coordinate in the observation space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        position = np.array([position_x, position_y])
        if isinstance(color, list):
            color = np.array(color)

        observation_space_range = (self.position_space.high - self.position_space.low)
        center = (position - self.position_space.low) / observation_space_range
        center[1] = 1 - center[1]
        center_y, center_x = (image.shape[:2] * np.flip(center)).astype(int)

        # Imagine a square of size width * width, with the coordinates computed above as a center. Iterate through
        # each pixel inside this square to
        radius = round(width / 2) + 1
        for i in range(center_x - radius, center_x + radius):
            for j in range(center_y - radius, center_y + radius):
                if distance.euclidean((i + 0.5, j + 0.5), (center_x, center_y)) < radius:
                    if 0 <= j < image.shape[0] and 0 <= i < image.shape[1]:
                        image[j, i] = color

        return image

    def place_edge(self, image: np.ndarray, position_1_x, position_1_y, position_2_x, position_2_y,
                   color: Union[np.ndarray, list]):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the observation space of the point to place.
        param y: y coordinate in the observation space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """

        position_1 = np.array([position_1_x, position_1_y])
        position_2 = np.array([position_2_x, position_2_y])

        color = np.array(color) if isinstance(color, list) else color
        observation_space_range = (self.position_space.high - self.position_space.low)

        center = (position_1 - self.position_space.low) / observation_space_range
        center[1] = 1 - center[1]
        center_y_1, center_x_1 = (image.shape[:2] * np.flip(center)).astype(int)

        center = (position_2 - self.position_space.low) / observation_space_range
        center[1] = 1 - center[1]
        center_y_2, center_x_2 = (image.shape[:2] * np.flip(center)).astype(int)

        rr, cc, val = line_aa(center_y_1, center_x_1, center_y_2, center_x_2)
        old = image[rr, cc]
        extended_val = np.tile(val, (3, 1)).T
        image[rr, cc] = (1 - extended_val) * old + extended_val * color

        return image


if __name__ == "__main__":
    # TODO
    # - Sampler des buts à une distance max
    test = True
    environment = DoomMaze(map_tag=DoomMapIndex.IMPOSSIBLE,
                           reset_anywhere=False,
                           discrete_goals=False,
                           goal_range=(15, 20),
                           test=test,
                           pixels_per_tile=5)

    results = []
    output_directory = os.path.dirname(os.path.abspath(__file__))
    for i in range(10):
        print("Episode #" + str(i + 1))
        images = []

        environment.reset()

        goal_image = environment.goal_image.copy()
        save_image(goal_image, output_directory + "/goals", "goal_episode_" + str(i))

        results.append(0)
        done = False
        for interaction_id in range(250):
            action = None if test else environment.action_space.sample()
            observation, reward, done = environment.step(action)

            # images.append(observation)
            image = environment.render(mode="top_view")
            images.append(image)
            if done:
                results[-1] = 1
                break
        print("results = ", results)
        generate_video(images, output_directory, "episode_" + str(i))

        print("Episode finished!")
        print("************************")
    a = 1



from enum import Enum


class Direction(Enum):
    TOP = 0
    BOTTOM = 1
    LEFT = 2
    RIGHT = 3


class Colors(Enum):
    EMPTY = [250, 250, 250]
    WALL = [50, 54, 51]
    START = [213, 219, 214]
    TRAP = [73, 179, 101]
    TILE_BORDER = [50, 54, 51]
    AGENT = [0, 0, 255]
    GOAL = [255, 0, 0]
    REWARD = [255, 215, 0]  # Added missing REWARD color (gold)


class TileType(Enum):
    EMPTY = 0
    WALL = 1
    START = 2
    REWARD = 3
    TRAP = 4


class PointMazeMapsIndex(Enum):
    EMPTY = "empty_room"
    FOUR_ROOMS = "four_rooms"
    MEDIUM = "medium_maze"
    JOIN_ROOM_MEDIUM = "join_room_medium"
    HARD = "hard_maze"
    EXTREME = "extreme_maze"
    IMPOSSIBLE = "impossible_maze"



import math
import random
import numpy as np
import networkx as nx
from typing import Union, Any, Optional
from gymnasium import Env
from gymnasium.spaces import Box
import mujoco_py
from scipy.spatial import distance
from skimage.draw import line_aa

from .mujoco_files.xml_generator import generate_xml
from .mujoco_model_utils import euler2quat
from enum import Enum


class TileType(Enum):
    EMPTY = 0
    WALL = 1
    START = 2
    REWARD = 3


def get_quaternion_from_euler(roll, pitch, yaw):
    """
    Convert an Euler angle to a quaternion.

    Input
      :param roll: The roll (rotation around x-axis) angle in radians.
      :param pitch: The pitch (rotation around y-axis) angle in radians.
      :param yaw: The yaw (rotation around z-axis) angle in radians.

    Output
      :return qx, qy, qz, qw: The orientation in quaternion [x,y,z,w] format
    """
    qx = np.sin(roll / 2) * np.cos(pitch / 2) * np.cos(yaw / 2) - np.cos(roll / 2) * np.sin(pitch / 2) * np.sin(yaw / 2)
    qy = np.cos(roll / 2) * np.sin(pitch / 2) * np.cos(yaw / 2) + np.sin(roll / 2) * np.cos(pitch / 2) * np.sin(yaw / 2)
    qz = np.cos(roll / 2) * np.cos(pitch / 2) * np.sin(yaw / 2) - np.sin(roll / 2) * np.sin(pitch / 2) * np.cos(yaw / 2)
    qw = np.cos(roll / 2) * np.cos(pitch / 2) * np.cos(yaw / 2) + np.sin(roll / 2) * np.sin(pitch / 2) * np.sin(yaw / 2)

    return [qx, qy, qz, qw]


class AntMazeV0(Env):

    name = "Visual Ant-Maze"

    def __init__(self, maze_name="empty_room", image_resolution_per_tile=50, show=False, random_orientation=False):
        """
        Initialise an ant maze environment.
        THe model is automatically created using a map specification defined in /mujoco_files/maps/
        The maze_name should correspond to a map name, so /mujoco_files/maps/<maze_name>.txt should exist.
        if random_orientation=True, env.reset() with reset the agent with a random orientation. Else, agent's orientation
        if 0.
        """
        self.random_orientation = random_orientation
        self.maze_name = maze_name
        self.image_resolution_per_tile = image_resolution_per_tile
        self.maze_array, xml_spec_path = generate_xml(maze_name)
        self.maze_array = np.array(self.maze_array)
        self.maze_array_height, self.maze_array_width = self.maze_array.shape

        # Create Mujoco Simulation
        self.model = mujoco_py.load_model_from_path(xml_spec_path)
        self.sim = mujoco_py.MjSim(self.model)

        # Get cameras
        # Create viewer for rendering (optional for visual debugging)
        self.viewer = mujoco_py.MjRenderContextOffscreen(self.sim, device_id=0)
        self.ant_view_camera_id = self.model.camera_name2id("ant_view")
        self.ant_render_camera_id = self.model.camera_name2id("ant_render")
        self.env_view_camera_id = self.model.camera_name2id("env_view")

        low = self.map_coordinates_to_env_position(0, -1)
        high = self.map_coordinates_to_env_position(-1, 0)
        self.maze_space = Box(low=np.float32(low), high=np.float32(high))  # Verify low and high
        self.goal_space = Box(low=np.float32(low), high=np.float32(high))
        self.goal_size = self.goal_space.shape[0]

        # Observation space
        self.observation_size = len(self.sim.data.qpos) + len(self.sim.data.qvel)
        fill_size = self.observation_size - self.maze_space.shape[0]
        observation_low = np.concatenate((self.maze_space.low, np.full(fill_size, float("-inf"))))
        observation_high = np.concatenate((self.maze_space.high, np.full(fill_size, float("inf"))))
        self.observation_space = Box(low=np.float32(observation_low), high=np.float32(observation_high))

        # Action space
        self.action_space = Box(low=np.float32(self.sim.model.actuator_ctrlrange[:, 0]),
                                high=np.float32(self.sim.model.actuator_ctrlrange[:, 1]))
        self.action_size = self.action_space.shape[0]

        max_actions = 500
        num_frames_skip = 10

        self.goal_thresholds = np.array([0.5, 0.5, 0.2, 0.5, 0.5])

        self.max_actions = max_actions

        # # Implement visualization if necessary
        self.num_frames_skip = num_frames_skip

        self.goal = None

    # Get observation, which concatenates joint positions and velocities
    def get_observation(self):
        return np.concatenate((self.sim.data.qpos, self.sim.data.qvel))

    def map_coordinates_to_env_position(self, x, y):
        x = self.maze_array_width + x if x < 0 else x
        y = self.maze_array_height + y if y < 0 else y
        return np.array([x - self.maze_array_width / 2 + 0.5, - (y - self.maze_array_height / 2 + 0.5)])

    def sample_reachable_position(self):
        # sample empty tile
        tile_coordinates = np.flip(random.choice(np.argwhere(self.maze_array != TileType.WALL.value)))
        position = self.map_coordinates_to_env_position(*tile_coordinates)

        # Sample a random position in the chosen tile
        low, high = position - 0.5, position + 0.5
        return np.random.uniform(low, high)

    # Reset simulation to observation within initial observation specified by user
    def reset(self, *,
              seed: int | None = None,
              options: dict[str, Any] | None = None, ):

        if seed is not None:
            random.seed(seed)

        # Reset controls
        self.sim.data.ctrl[:] = 0

        # Reset joint positions and velocities

        # Sample a start position
        start_tiles = np.argwhere(self.maze_array == TileType.START.value)
        start_coordinates = np.flip(random.choice(start_tiles))
        start_position = self.map_coordinates_to_env_position(*start_coordinates)

        # Build the observation from it, using fixed joints and velocities values.
        orientation = random.random() * 2 * math.pi if self.random_orientation else 0
        quaternion_angle = get_quaternion_from_euler(orientation, 0, 0)
        initial_qpos = np.array([0.55] + quaternion_angle + [0.0, 1.0, 0.0, -1.0, 0.0, -1.0, 0.0, 1.0])
        # initial_qpos = np.array([0.55, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, -1.0, 0.0, -1.0, 0.0, 1.0])
        initial_qvel = np.zeros(14)
        initial_observation = np.concatenate((start_position, initial_qpos, initial_qvel))

        self.sim.data.qpos[:] = initial_observation[:len(self.sim.data.qpos)]
        self.sim.data.qvel[:] = initial_observation[len(self.sim.data.qpos):]
        self.sim.step()

        # Choose a goal. Weights are used to keep a uniform sampling over the reachable space,
        # since boxes don't cover areas of the same size.
        self.goal = self.sample_reachable_position().copy()

        # Return observation
        observation = self.get_observation()
        return observation, {"goal": self.goal}

    # Execute low-level action for number of frames specified by num_frames_skip
    def step(self, action):
        self.sim.data.ctrl[:] = action
        for _ in range(self.num_frames_skip):
            self.sim.step()
        new_observation = self.get_observation()

        reached = (np.abs(new_observation[:len(self.goal)] - self.goal) < self.goal_thresholds[:len(self.goal)]).all()
        return self.get_observation(), 0 if reached else -1, reached, False, {"reached": reached}

    def update_graph(self, graph: nx.DiGraph):
        for node_id, node_attributes in graph.nodes(data=True):
            self.set_node("node_" + str(node_id), rgba="0 1 0 1", position=node_attributes["observation"])

        for edge_id, (node_1, node_2, edge_attributes) in enumerate(graph.edges(data=True)):
            first_node = nx.get_node_attributes(graph, "observation")[node_1]
            second_node = nx.get_node_attributes(graph, "observation")[node_2]
            self.set_edge("edge_" + str(edge_id), first_node, second_node, "0 1 0 1")

    # Visualize end goal. This function may need to be adjusted for new environments.
    def display_end_goal(self):
        self.sim.data.mocap_pos[0][:2] = np.copy(self.goal[:2])

    # Visualize all sub-goals
    def display_sub_goals(self, sub_goals):

        # Display up to 10 sub-goals and end goal
        if len(sub_goals) <= 11:
            sub_goal_ind = 0
        else:
            sub_goal_ind = len(sub_goals) - 11

        for i in range(1, min(len(sub_goals), 11)):
            self.sim.data.mocap_pos[i][:2] = np.copy(sub_goals[sub_goal_ind][:2])
            self.sim.model.site_rgba[i][3] = 1

            sub_goal_ind += 1

    def set_node(self, node_name, rgba=None, position=None):
        geom_id = self.sim.model.geom_name2id(node_name)

        if rgba is not None:
            if isinstance(rgba, str):
                rgba = np.array([float(elt) for elt in rgba.split(" ")])
            if isinstance(rgba, list):
                rgba = np.array(rgba)
            self.sim.model.geom_rgba[geom_id, :] = rgba

        if position is not None:
            if isinstance(position, list):
                position = np.array(position)
            assert len(position) <= 3
            self.sim.model.geom_pos[geom_id, :len(position)] = position

    def set_edge(self, edge_name, first_node: np.ndarray, second_node: np.ndarray, rgba=None):
        edge_id = self.sim.model.geom_name2id(edge_name)


        # Compute edge position
        position = (first_node + second_node) / 2
        self.sim.model.geom_pos[edge_id] = position

        # Compute angle between these two nodes
        diff = second_node - first_node
        angle = math.acos(diff[0] / np.linalg.norm(diff))
        euler_rotation = np.array([-math.pi / 2, angle, 0])
        self.sim.model.geom_quat[edge_id] = euler2quat(euler_rotation)

        if rgba is not None:
            if isinstance(rgba, str):
                rgba = np.array([float(elt) for elt in rgba.split(" ")])
            if isinstance(rgba, list):
                rgba = np.array(rgba)
            self.sim.model.geom_rgba[edge_id, :] = rgba

    def toggle_geom_object(self, object_name):
        geom_id = self.sim.model.geom_name2id(object_name)
        self.sim.model.geom_rgba[geom_id, -1] = 1 - self.sim.model.geom_rgba[geom_id, -1]

    def set_space_color(self, image_array: np.ndarray, low, high, color) -> np.ndarray:
        """
        Set a tile color with the given color in the given image as a numpy array of pixels
        :param image_array: The image where the tile should be set
        :param low: the box lower corner
        :param high: the box higher corner
        :param color: new color of the tile : numpy array [Red, Green, Blue]
        :return: The new image
        """
        tile_size = self.image_resolution_per_tile
        tile_width, tile_height = np.rint((high - low) * tile_size).astype(int)
        tile_img = np.tile(color, (tile_height, tile_width, 1))

        image_width, image_height, _ = image_array.shape

        # We add +1 to take images borders in consideration. Without it, x/y_min positions are computed from the
        # border of self.maze_space, that do not include maze borders.
        # NB: Y coordinates don't work the same in the observation space and in the image.
        # High values (in the observation space) are at the top of the image (low numpy index), with low coordinates in
        # the image. Opposite for X values.
        x_min = round((low[0] - self.maze_space.low[0] + 1) * tile_size)
        y_max = round(((image_height / tile_size) - (low[1] - self.maze_space.low[1] + 1)) * tile_size)

        x_max = x_min + tile_width
        y_min = y_max - tile_height

        image_array[y_min:y_max, x_min:x_max, :] = tile_img
        return image_array

    def get_goal_from_observation(self, observation: np.ndarray) -> np.ndarray:
        return observation.copy()[:self.goal_size]

    def get_depth_image(self, width, height, camera_id):
        self.viewer.render(width, height, camera_id=camera_id)
        _, depth = self.viewer.read_pixels(width, height, depth=True)
        depth = depth[::-1, :]
        # print("Raw depth buffer range:", depth.min(), depth.max())
        near = self.model.vis.map.znear
        far = self.model.vis.map.zfar
        # print("znear:", near, "zfar:", far)
        # depth_buffer ∈ [0, 1]
        z = depth * 2.0 - 1.0  # Convert to [-1, 1]
        real_depth = 2.0 * near * far / (far + near - z * (far - near))
        # print("Real depth range:", real_depth.min(), real_depth.max())

        # Optional: normalize for visualization
        norm = (real_depth - real_depth.min()) / (real_depth.max() - real_depth.min() + 1e-8)
        depth_img = (255 * (1.0 - norm)).astype(np.uint8)  # Invert so near = bright
        return depth_img

    def render(self, mode="full", ant_color: Optional[tuple] = None, ignore_goal=False) -> np.ndarray:
        """
        Return a np.ndarray of size  of pixels that represent the environments and it's walls
        :param mode: rendering mode, possible values:
            - "generated": generate an image with a white background, and where walls are added in black.
            Points of various colors are also added to represent the agent and goals positions.
            This mode is ideal to capture the
            - "full" [default]: get an image from a camera placed on top of the environment.
            It is able to capture the entire maze. Ideal to visualise the entire path made by the ant.
            - "ant": get an image from a camera placed on top of the ant.
            This camera follow the ant. Ideal to visualise the entire path made by the ant.
        :param ignore_goal: if false, does not render the goals.
        :param ant_color:
        :return: The final image.
        """

        if ant_color is not None:
            raise NotImplementedError("")

        if mode == "generated":

            # we add np.full(2, 2) to get image borders, because border walls are not included in env.maze_space.
            # We assume walls have a width of one, but it's not really important for a generated image.
            image_width_px, image_height_px = np.rint((self.maze_space.high - self.maze_space.low + np.full(2, 2))
                                                      * self.image_resolution_per_tile).astype(int)

            img = np.zeros(shape=(image_height_px, image_width_px, 3), dtype=np.uint8)

            # Render the grid
            for coordinates in np.argwhere(self.maze_array != TileType.WALL.value):
                coordinates = np.flip(coordinates)
                position = self.map_coordinates_to_env_position(*coordinates)
                img = self.set_space_color(img, position - 0.5, position + 0.5, np.array([255, 255, 255]))

            self.place_point(img, self.sim.data.qpos[:2], [0, 0, 255], width=10)
            if not ignore_goal:
                if self.goal is None:
                    raise Exception("The environment have no goal to print, please reset the environment using .reset() before to call .render().")
                self.place_point(img, self.goal[:2], [255, 0, 0], width=10)
            return img
        if mode == "full":
            # Get the image from the full cam
            width, height = 640, 480
            return self.get_depth_image(width, height, camera_id=self.env_view_camera_id)
            # self.viewer.render(width, height, camera_id=self.env_view_camera_id)
            # return self.viewer.read_pixels(width, height, depth=True)[1] * 255
        if mode == "ant":
            # Get the image from the camera on top of the ant
            width, height = 640, 480
            return self.get_depth_image(width, height, camera_id=self.ant_render_camera_id)
            # self.viewer.render()
            # return self.viewer.read_pixels(width, height, depth=True)[1] * 255
        if mode == "pov":
            # Get the image from the camera on top of the ant
            width, height = 640, 480
            return self.get_depth_image(width, height, camera_id=self.ant_view_camera_id)
            # self.viewer.render(width, height, camera_id=self.ant_view_camera_id)
            # return self.viewer.read_pixels(width, height, depth=True)[1] * 255
        else:
            raise NotImplementedError("")

    def place_point(self, image: np.ndarray, observation, color: Union[np.ndarray, list], width=5):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the observation space of the point to place.
        param y: y coordinate in the observation space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        if isinstance(color, list):
            color = np.array(color)

        observation_space_range = (self.observation_space.high[:2] - self.observation_space.low[:2])
        center = (observation[:2] - self.observation_space.low[:2]) / observation_space_range
        center[1] = 1 - center[1]
        center_y, center_x = (image.shape[:2] * np.flip(center)).astype(int)

        # Imagine a square of size width * width, with the coordinates computed above as a center. Iterate through
        # each pixel inside this square to
        radius = width
        for i in range(center_x - radius, center_x + radius):
            for j in range(center_y - radius, center_y + radius):
                dist = distance.euclidean((i, j), (center_x, center_y))
                if dist < radius and 0 <= i < image.shape[1] and 0 <= j < image.shape[0]:
                    image[j, i] = color
        return image

    def place_edge(self, image: np.ndarray, observation_1, observation_2, color: Union[np.ndarray, list], width=40):
        """
        Modify the input image
        param image: Initial image that will be modified.
        param x: x coordinate in the observation space of the point to place.
        param y: y coordinate in the observation space of the point to place.
        param color: Color to give to the pixels that compose the point.
        param width: Width of the circle (in pixels).
        """
        if isinstance(color, list):
            color = np.array(color)
        observation_space_range = (self.observation_space.high[:2] - self.observation_space.low[:2])

        center = (observation_1[:2] - self.observation_space.low[:2]) / observation_space_range
        center[1] = 1 - center[1]
        center_y_1, center_x_1 = (image.shape[:2] * np.flip(center)).astype(int)

        center = (observation_2[:2] - self.observation_space.low[:2]) / observation_space_range[:2]
        center[1] = 1 - center[1]
        center_y_2, center_x_2 = (image.shape[:2] * np.flip(center)).astype(int)

        rr, cc, val = line_aa(center_y_1, center_x_1, center_y_2, center_x_2)
        old = image[rr, cc]
        extended_val = np.tile(val, (3, 1)).T
        image[rr, cc] = (1 - extended_val) * old + extended_val * color



import copy
import importlib
import math
import random
from enum import Enum
from typing import Tuple, Optional, Any
import numpy as np
from gymnasium import spaces, Env


class GridWorldMapsIndex(Enum):
    EMPTY = "empty_room"
    FOUR_ROOMS = "four_rooms" 
    MEDIUM = "medium_maze"
    HARD = "hard_maze"
    EXTREME = "extreme_maze"


class Colors(Enum):
    EMPTY = [250, 250, 250]
    WALL = [50, 54, 51]
    START = [213, 219, 214]
    TRAP = [255, 0, 0]
    TILE_BORDER = [50, 54, 51]
    AGENT = [0, 0, 255]
    GOAL = [73, 179, 101]


class TileType(Enum):
    EMPTY = 0
    WALL = 1
    START = 2
    REWARD = 3
    TRAP = 4


class Directions(Enum):
    TOP = 0
    BOTTOM = 1
    LEFT = 2
    RIGHT = 3


class GridWorldV0(Env):
    def __init__(self, **params):
        self.map_name: str = params.get("map_name", GridWorldMapsIndex.EMPTY.value)
        self.goal_conditioned: bool = params.get("goal_conditioned", True)
        self.reset_anywhere: bool = params.get("reset_anywhere", True)
        self.stochasticity: float = params.get("stochasticity", 0.0)
        assert 0.0 <= self.stochasticity <= 1.0, f"Invalid stochasticity value: {self.stochasticity}"

        self.goal_position: Optional[Tuple[int, int]] = None
        self.agent_coordinates: Optional[Tuple[int, int]] = None

        self.maze_array = np.array(importlib.import_module(f"rlnav.grid_world.maps.{self.map_name}").maze_array)
        self.height, self.width = self.maze_array.shape

        self.observation_space = spaces.Discrete(self.height * self.width)
        self.action_space = spaces.Discrete(len(Directions))

        self.reset()

    def reset(self, *,
              seed: int | None = None,
              options: dict[str, Any] | None = None, ):

        if seed is not None:
            random.seed(seed)

        if self.goal_conditioned:
            self.goal_position = self.sample_reachable_position()
            assert self.goal_position is not None, "No valid goal positions found."

        start_positions = np.argwhere(self.maze_array == TileType.START.value)
        assert len(start_positions) > 0, "No valid start positions found."
        self.agent_coordinates = tuple(random.choice(start_positions))

        return self._get_observation(*self.agent_coordinates), {"goal": self.goal} if self.goal_conditioned else {}

    def step(self, action: int):
        info = {
            "agent_coordinates": self.agent_coordinates,
            "agent_new_coordinates": self.agent_coordinates,
        }
        if self.goal_conditioned:
            info["goal"] = self.goal.copy()

        action = self._apply_stochasticity(action)
        new_i, new_j = self._compute_new_position(action)

        if not self._is_valid_move(new_i, new_j):
            return self._get_observation(*self.agent_coordinates), 0, False, False, info

        tile_type = self.get_tile_type(new_i, new_j)
        reward = -1 if tile_type == TileType.TRAP else 0
        terminated = tile_type in {TileType.REWARD, TileType.TRAP}

        if self.goal_conditioned:
            if (new_i, new_j) == self.goal_position:
                reward = 0
                terminated = True

        self.agent_coordinates = (new_i, new_j)
        info["agent_new_coordinates"] = self.agent_coordinates
        return self._get_observation(new_i, new_j), reward, terminated, False, info

    def _apply_stochasticity(self, action: int) -> int:
        if self.stochasticity > 0.0 and random.random() < self.stochasticity:
            action = (action + random.choice([-1, 1])) % 4
        return action

    def _compute_new_position(self, action: int) -> Tuple[int, int]:
        i, j = self.agent_coordinates
        if action == Directions.TOP.value and i > 0:
            i -= 1
        elif action == Directions.BOTTOM.value and i < self.height - 1:
            i += 1
        elif action == Directions.LEFT.value and j > 0:
            j -= 1
        elif action == Directions.RIGHT.value and j < self.width - 1:
            j += 1
        return i, j

    def _is_valid_move(self, i: int, j: int) -> bool:
        return 0 <= i < self.width and 0 <= j < self.height and self.get_tile_type(i, j) != TileType.WALL

    def _get_observation(self, i: int, j: int) -> int:
        return i * self.width + j

    def get_tile_type(self, i: int, j: int) -> TileType:
        return TileType(self.maze_array[i, j])

    def sample_reachable_position(self) -> Optional[Tuple[int, int]]:
        empty_positions = np.argwhere(
            np.logical_or(self.maze_array == TileType.EMPTY.value, self.maze_array == TileType.REWARD.value))
        if len(empty_positions) == 0:
            raise ValueError("No valid goal positions found.")
        return tuple(random.choice(empty_positions))

    def is_available(self, i: int, j: int) -> bool:
        return self._is_valid_move(i, j)
    
    def render(self, show_agent=True, show_rewards=True):
        pixels_per_tile = 10
        def set_tile_color(image_, i, j, color):
            image_[expanded_array == TileType.REWARD.value] = Colors.EMPTY.value
            i_start, i_end = i * pixels_per_tile, (i + 1) * pixels_per_tile
            j_start, j_end = j * pixels_per_tile, (j + 1) * pixels_per_tile
            image_[i_start:i_end, j_start:j_end, :] = color
            return image_

        # Second trial
        expanded_array = np.kron(self.maze_array, np.ones((pixels_per_tile, pixels_per_tile), dtype=int))
        image = np.zeros(expanded_array.shape + (3,), dtype=np.uint8)
        image[expanded_array == TileType.WALL.value] = Colors.WALL.value
        image[expanded_array == TileType.EMPTY.value] = Colors.EMPTY.value
        image[expanded_array == TileType.START.value] = Colors.START.value
        if show_rewards:
            image[expanded_array == TileType.TRAP.value] = Colors.TRAP.value
            if self.goal_conditioned:
                image = set_tile_color(image, *self.goal_position, Colors.GOAL.value)
            else:
                image[expanded_array == TileType.REWARD.value] = Colors.GOAL.value
        if show_agent:
            image = set_tile_color(image, *self.agent_coordinates, Colors.AGENT.value)
        return image
                
    def copy(self):
        return copy.deepcopy(self)
